require 'game'

function load()
    input = Input()
    timer = Timer()
    time = 0
    frame = 0

    -- Input
    input:bind('f1', function()
        Lurker.scan()
        requireAll()
        engineStart()
    end)

    -- Misc 
    camera = Camera(gw/2, gh/2)
    slow_amount = 1
    flash_color = {0, 0, 0, 255}

    -- Graphics
    love.graphics.setLineStyle('rough')

    -- Start
    engineStart()
end

function engineStart()
    rooms = {}
    current_room = nil
    gameStart()
end

function update(dt)
    time = time + dt
    frame = frame + 1
    timer:update(dt*slow_amount)
    camera:update(dt*slow_amount)
    if current_room then current_room:update(dt*slow_amount) end
end

function draw()
    if current_room then current_room:draw() end

    if flash_frames then 
        flash_frames = flash_frames - 1
        if flash_frames == -1 then flash_frames = nil end
    end
    if flash_frames then
        love.graphics.setColor(flash_color)
        love.graphics.rectangle('fill', 0, 0, sx*gw, sy*gh)
        love.graphics.setColor(255, 255, 255)
    end
end

function resize(s)
    love.window.setMode(s*gw, s*gh)
    sx, sy = s, s
end

function quit()
    endSession()
    sendSessionToServer()
end



-- Effects --
function flash(frames, color)
    flash_frames = frames
    flash_color = color or {0, 0, 0, 255}
end

function slow(amount, duration)
    slow_amount = amount
    timer:tween('slow', duration, _G, {slow_amount = 1}, 'in-out-cubic')
end



-- Room Management --
function addRoom(room_type)
    local room = _G[room_type]()
    table.insert(rooms, room)
    return room
end

function getRoom(room_id)
    local target_room
    for _, room in ipairs(rooms) do 
        if room.id == room_id then 
            target_room = room
            break
        end
    end
    return target_room
end

function gotoRoom(room)
    if current_room then removeRoom(current_room.id) end
    current_room = addRoom(room)
end

function removeRoom(room_id)
    for i, room in ipairs(rooms) do
        if room.id == room_id then
            table.remove(rooms, i)
        end
    end
end



-- Load, Require --
function loadFonts(path)
    fonts = {}
    local font_paths = {}
    recursiveEnumerate(path, font_paths)
    for i = 8, 64, 4 do
        for _, font_path in pairs(font_paths) do
            local last_forward_slash_index = font_path:find("/[^/]*$")
            local font_name = font_path:sub(last_forward_slash_index+1, -5)
            local font = love.graphics.newFont(font_path, i)
            font:setFilter(filter_mode, filter_mode, anisotropy)
            fonts[font_name .. '_' .. i] = font
        end
    end
end

function loadGraphics(path)
    assets = {}
    local asset_paths = {}
    recursiveEnumerate(path, asset_paths)
    for _, asset_path in pairs(asset_paths) do
        local last_forward_slash_index = asset_path:find("/[^/]*$")
        local asset_name = asset_path:sub(last_forward_slash_index+1, -5)
        local image = love.graphics.newImage(asset_path, {mipmaps = true})
        image:setFilter(filter_mode, filter_mode, anisotropy)
        assets[asset_name] = image
    end
end

function loadShaders(path)
    shaders = {}
    local shader_paths = {}
    recursiveEnumerate(path, shader_paths)
    for _, shader_path in pairs(shader_paths) do
        local last_forward_slash_index = shader_path:find("/[^/]*$")
        local shader_name = shader_path:sub(last_forward_slash_index+1, -6)
        local shader = love.graphics.newShader(shader_path)
        shaders[shader_name] = shader
    end
end

function recursiveEnumerate(folder, file_list)
    local items = love.filesystem.getDirectoryItems(folder)
    for _, item in ipairs(items) do
        local file = folder .. '/' .. item
        if love.filesystem.isFile(file) then
            table.insert(file_list, file)
        elseif love.filesystem.isDirectory(file) then
            recursiveEnumerate(file, file_list)
        end
    end
end

function requireAll()
    loadFonts('resources/fonts')
    loadGraphics('resources/graphics')
    loadShaders('resources/shaders')

    local object_files = {}
    recursiveEnumerate('objects', object_files)
    requireFiles(object_files)

    local room_files = {}
    recursiveEnumerate('rooms', room_files)
    requireFiles(room_files)
end

function requireFiles(files)
    for _, file in ipairs(files) do
        local last_forward_slash_index = file:find("/[^/]*$")
        local name = file:sub(last_forward_slash_index+1, -5)
        package.loaded[file:sub(1, -5)] = nil
        _G[name] = require(file:sub(1, -5))
    end
end



-- Session -- 
function setUUID()
    if love.filesystem.exists('uuid') then
        uuid = love.filesystem.read('uuid')

    -- Create new UUID for this user if its the first time
    else
        uuid = Lume.uuid()
        love.filesystem.write('uuid', uuid) 
    end
end

function setSession()
    if love.filesystem.exists('session') then
        session = love.filesystem.read('session')
        session = session + 1

    -- Start session at 1 if its the first time
    else
        session = 1
        love.filesystem.write('session', session)
    end
end

function startSession()
    setUUID()
    setSession()

    session_data = {}
    session_data.uuid = uuid
    session_data.session = session
    session_data.session_start_date = os.date("*t")
    init_time = os.time()

    session_data.events = {}
end

function endSession()
    love.filesystem.write('session', session)

    session_data.session_duration = os.difftime(os.time(), init_time)
    session_data.session_end_date = os.date("*t")

    -- Save session_data
    local session_data_string = Serialize(session_data)
    if love.filesystem.exists('sessions') then
        love.filesystem.write('sessions/' .. session, session_data_string)
    else
        love.filesystem.createDirectory('sessions')
        love.filesystem.write('sessions/' .. session, session_data_string)
    end
end

function sendSessionToServer()
    -- Send session_data_string to server
    local session_data_string = Serialize(session_data)
    local host = enet.host_create()
    local server = host:connect("104.131.70.183:6789")
    local done = false
    local sent_to_server = false
    while not done do
        local event = host:service(1000)
        if event then
            if event.type == "connect" then event.peer:send(session_data_string)
            elseif event.type == "receive" then done = true; sent_to_server = true end
        -- Timeout
        else done = true end
    end
    server:disconnect()
    host:flush()

    session_data.sent_to_server = sent_to_server
end
