--- Sound v1.3, a simplified sound system for Love 2D
-- @author Ensayia (Ensayia@gmail.com) & Taehl (SelfMadeSpirit@gmail.com)
-- @author SiENcE^Florian Fischer ... edited to support Source & SoundData and global options to enable & disable play (sounds) or playlooping (music)

--It's under the ZLIB license.

-- Sound system created by Ensayia and expanded by Taehl
Sound = {}				-- Namespace
Sound.channels = {}		-- This holds the currently playing sound channels
Sound.volumeLevels = {}	-- Volume levels that multiply the volumes of sounds with those tags
Sound.pitchLevels = {}	-- Pitch levels that multiply the pitches of sounds with those tags

Sound.soundsPlayedThisFrame = {}
Sound.currentlyPlayingSources = 0
Sound.currentlyPlayingSourcesLimit = 32
Sound.sound = false
Sound.music = false 

-- Functions for playing sounds

--- Play a sound, or a random sound from a list. (with optional tag(s), volume, pitch, and on-finished function)
-- @param sound A filepath to the sound file (example: "sounds/jump.ogg"), or a list of filepaths. If a list is used, a random sound from that list will be played. Passing a string is shorthand for passing a table with only one item.
-- @param tags One or more tags that can be used to identify a sound (Example, single: "music". Example, multiple: {"sfx", "gun", "player"}). Multiple sounds may share tags, or tags may be unique. If omitted, no tags are assigned. Passing a string is shorthand for passing a table with only one item.
-- @param volume A number between 0 and 1 which specifies how loud a sound is. If the sound has a tag which a volume has been specified for, it will multiply this number (ie., if you use the tag "sfx" and volume 0.5, and "sfx" has been tagVolume-ed to 0.6, then the sound will be played at 0.3 volume).
-- @param pitch A number which specifies the speed/pitch the sound will be played it. If the sound has a tag which a pitch has been specified for, it will multiply this number.
-- @param func A function which will be called when the sound is finished playing (it's passed one parameter - a list with the sound's volume and pitch). If omitted, no function will be used.
function Sound.play(sound, tags, volume, pitch, func)
	if not Sound.sound then return end

	if not sound then
		print("Your sound source is nil.")
		return nil, "Your sound source is nil."
	end
	
	-- arrays not supported, because loading of sounds had to be done before
	-- if type(sound) == "table" then sound = sounds[sound[math.random(#sound)]] 
    -- else sound = sounds[sound] end
	
	if sound.type and sound:type() ~= "Source" and sound:type() ~= "SoundData" then
		print("You must use a valid sound source (Source or SoundData).")
		return nil, "You must use a valid sound source."
	end
		
	if Sound.soundsPlayedThisFrame[sound] then return end	
	if Sound.currentlyPlayingSources >= Sound.currentlyPlayingSourcesLimit then return end

	Sound.soundsPlayedThisFrame[sound] = true
	Sound.currentlyPlayingSources = Sound.currentlyPlayingSources + 1
		
	if sound.type and sound:type() == "Source" then
		table.insert(Sound.channels, { sound, func, {volume or 1, pitch or 1}, tags=(type(tags) == "table" and tags or {tags}) })
--		print( "play music - channel:", #Sound.channels, tags)
	else
		table.insert(Sound.channels, { love.audio.newSource(sound), func, {volume or 1, pitch or 1}, tags=(type(tags) == "table" and tags or {tags}) })
--		print( "play sound - channel:", #Sound.channels, tags )
--		print( "play sound - channel:", #Sound.channels, tags, sound:getChannels(), sound:getSampleRate() )
	end

	local s = Sound.channels[#Sound.channels]
	s[1]:play()
	s[1]:setVolume( (volume or 1) * Sound.findVolume(tags) * (Sound.volumeLevels.all or 1) )
	s[1]:setPitch( (pitch or 1) * Sound.findPitch(tags) * (Sound.pitchLevels.all or 1) )

	return #Sound.channels
end

--[[
-- play a 3D sound - you have to set the listener in your game code!!
-- returns the source for manual stop/pause aso.
function Sound.play3d(sound, tags, volume, pitch, func, x, y, z, distmin, distmax)
	local source = love.audio.newSource( sound )
	source:setDistance( distmin or 1, distmax or 100 )
	source:setPosition( x or 0, y or 0, z or 0 )
	Sound.play(source, tags, volume, pitch, func)
	return source
end
]]--

--- Play a repeating sound, or random sounds from a list. (with optional tag(s), volume, and pitch)
-- @param sound See Sound.play
-- @param tags See Sound.play
-- @param n The number of times the sound will be looped. If omitted, the sound will loop until Sound.stop is used.
-- @param volume See Sound.play
-- @param pitch See Sound.play
function Sound.playLooping(sound, tags, n, volume, pitch)
	if not Sound.music then return end
	return Sound.play( sound, tags, volume, pitch,
		(not n or n > 1) and function(d) Sound.playLooping(sound, tags, (n and n-1), d[1], d[2]) end
	)
end


-- Functions for modifying sounds that are playing (passing these a tag instead of a string is generally preferable)

--- Sets the volume of channel or tag and its loops (if any).
-- Volume/pitch changes take effect immediately and last until changed again. Sound "tags" are recommended for changing multiple sounds together, so you can independently adjust the volume of all sound effects, music, etc. Just don't forget to tag your sounds! If a sound has multiple tags with set volumes, the first one (in the order of its tag list) will be used.
-- Example: Sound.volume("music", .5))	-- set music to half volume
-- @param channel Determines which channel (number) or tag (string) will be affected. Since sound channels aren't static, it's advisable to use tags. The special sound tag "all" will always affect all sounds.
-- @param volume See Sound.play. If omitted, effectively resets volume to 1.
function Sound.volume(channel, volume)
	if type(channel) == "number" then
		local c = Sound.channels[channel] volume = volume or c[3][1] c[3][1] = volume
		c[1]:setVolume( volume * Sound.findVolume(c.tags) * (Sound.volumeLevels.all or 1) )
	elseif type(channel) == "string" then Sound.volumeLevels[channel]=volume for k,v in pairs(Sound.findTag(channel)) do Sound.volume(v, volume) end
	end
end

--- Sets the pitch of channel or tag and its loops (if any).
-- @param channel See Sound.volume
-- @param pitch See Sound.play. If omitted, effectively resets pitch to 1.
function Sound.pitch(channel, pitch)
	if type(channel) == "number" then
		local c = Sound.channels[channel] pitch = pitch or c[3][2] c[3][2] = pitch
		c[1]:setPitch( pitch * Sound.findPitch(c.tags) * (Sound.pitchLevels.all or 1) )
	elseif type(channel) == "string" then Sound.pitchLevels[channel]=pitch for k,v in pairs(Sound.findTag(channel)) do Sound.pitch(v, pitch) end
	end
end

--- Pauses a channel or tag. Use Sound.resume to unpause.
-- @param channel See Sound.volume
function Sound.pause(channel)
	if type(channel) == "number" then Sound.channels[channel][1]:pause()
	elseif type(channel) == "string" then for k,v in pairs(Sound.findTag(channel)) do Sound.pause(v) end
	end
end

--- Resumes a channel or tag from a pause.
-- @param channel See Sound.volume
function Sound.resume(channel)
	if type(channel) == "number" then Sound.channels[channel][1]:resume()
	elseif type(channel) == "string" then for k,v in pairs(Sound.findTag(channel)) do Sound.resume(v) end
	end
end

-- Stops a sound channel or tag either immediately or when finished and prevents it from looping.
-- @param channel See Sound.volume
-- @param finish If true, the sound will be allowed to finish playing instead of stopping immediately.
function Sound.stop(channel, finish)
	if type(channel) == "number" then local c = Sound.channels[channel] c[2] = nil if not finish then c[1]:stop() end
	elseif type(channel) == "string" then for k,v in pairs(Sound.findTag(channel)) do Sound.stop(v, finish) end
	end
end


-- Utility functions

--- Cleans up finished sounds, freeing memory (highly recommended to go in love.update()). If not called, memory and channels won't be freed, and sounds won't loop.
function Sound.cleanup()
	for k,v in ipairs(Sound.channels) do
		if v[1]:isStopped() then
			if v[2] then v[2](v[3]) end		-- allow sounds to use custom functions (primarily for looping, but be creative!)
			table.remove(Sound.channels, k)
			Sound.currentlyPlayingSources = Sound.currentlyPlayingSources - 1
		end
	end
	
	Sound.soundsPlayedThisFrame = {}
end

-- Add or change a default volume level for a specified tag (for example, to change music volume, use Sound.tagVolume("music", .5))
function Sound.tagVolume(tag, volume)
	Sound.volumeLevels[tag] = volume
	Sound.volume(tag)
end

-- Add or change a default pitch level for a specified tag
function Sound.tagPitch(tag, pitch)
	Sound.pitchLevels[tag] = pitch
	Sound.pitch(tag)
end


-- Internal functions

-- Returns a list of all sound channels with a given tag.
-- @param tag The channels of all sounds with this tag will be returned.
-- @return A list of sound channel numbers.
function Sound.findTag(tag)
	local t = {}
	for channel,sound in ipairs(Sound.channels) do
		if sound.tags then for k,v in ipairs(sound.tags) do
			if tag == "all" or v == tag then table.insert(t, channel) end
		end end
	end
	return t
end

--- Returns a volume level for a given tag or tags, or 1 if the tag(s)'s volume hasn't been set. If a list of tags is given, it will return the level of the first tag with a set volume.
-- @param tag Chooses the tag to check the volume of.
-- @return volume, a number between 0 and 1 (if nil, defaults to 1).
function Sound.findVolume(tag)
	if type(tag) == "string" then return Sound.volumeLevels[tag] or 1
	elseif type(tag) == "table" then for k,v in ipairs(tag) do if Sound.volumeLevels[v] then return Sound.volumeLevels[v] end end
	end
	return 1	-- if nothing is found, default to 1
end

--- Returns a pitch level for a given tag or tags, or 1 if the tag(s)'s pitch hasn't been set.
-- @param tag See Sound.findVolume
-- @return pitch, a number between 0 and 1 (if nil, defaults to 1).
function Sound.findPitch(tag)
	if type(tag) == "string" then return Sound.pitchLevels[tag] or 1
	elseif type(tag) == "table" then for k,v in ipairs(tag) do if Sound.pitchLevels[v] then return Sound.pitchLevels[v] end end
	end
	return 1	-- if nothing is found, default to 1
end
