function pushRotate(x, y, r)
    love.graphics.push()
    love.graphics.translate(x, y)
    love.graphics.rotate(r or 0)
    love.graphics.translate(-x, -y)
end

function pushRotateScale(x, y, r, sx, sy)
    love.graphics.push()
    love.graphics.translate(x, y)
    love.graphics.rotate(r or 0)
    love.graphics.scale(sx or 1, sy or sx or 1)
    love.graphics.translate(-x, -y)
end

function table.tostring(t)
    local str = ''
    for _, c in ipairs(t) do str = str .. c end
    return str
end

function table.random(t)
    return t[love.math.random(1, #t)]
end

function table.copy(t)
    local copy
    if type(t) == 'table' then
        copy = {}
        for k, v in next, t, nil do copy[table.copy(k)] = table.copy(v) end
        setmetatable(copy, table.copy(getmetatable(t)))
    else copy = t end
    return copy
end

function table.merge(t1, t2)
    local new_table = {}
    for k, v in pairs(t2) do new_table[k] = v end
    for k, v in pairs(t1) do new_table[k] = v end
    return new_table
end

function distance(x1, y1, x2, y2)
    return math.sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2))
end

function math.randomp(min, max)
    return (min > max and (love.math.random()*(min - max) + max)) or (love.math.random()*(max - min) + min)
end

function math.roundp(n, p)
    local m = math.pow(10, p or 0)
    return math.floor(n*m+0.5)/m
end

function math.roundm(n, m)
    return math.floor(((n + m - 1)/m))*m
end

function math.randomCircle(r)
    local angle = math.randomp(0, 1)*2*math.pi
    local radius = math.randomp(0, 1)*r
    return radius*math.cos(angle), radius*math.sin(angle)
end

function math.rangec(old_value, old_min, old_max, new_min, new_max)
    local new_min = new_min or 0
    local new_max = new_max or 1
    local new_value = 0
    local old_range = old_max - old_min
    if old_range == 0 then new_value = new_min
    else
        local new_range = new_max - new_min
        new_value = (((old_value - old_min)*new_range)/old_range) + new_min
    end
    return new_value
end

function angleToDirection2(angle)
    local angle = math.abs(angle)
    if angle < math.pi/2 or angle >= 3*math.pi/2 then return 'right' end
    if angle >= math.pi/2 and angle < 3*math.pi/2 then return 'left' end
end

function angleToDirection4(angle)
    local pi = math.pi
    if angle >= pi/4 then angle = -2*pi+angle end
    if angle < pi/4 and angle >= -1*pi/4 then return 'right' end
    if angle < -1*pi/4 and angle >= -3*pi/4 then return 'up' end
    if angle < -3*pi/4 and angle >= -5*pi/4 then return 'left' end
    if angle < -5*pi/4 and angle >= -7*pi/4 then return 'down' end
end

function directionToAngle4(dir)
    if dir == 'right' then return 0 end
    if dir == 'up' then return -math.pi/2 end
    if dir == 'left' then return math.pi end
    if dir == 'down' then return math.pi/2 end
end

function rectangleToVertexList(x, y, w, h)
    return {x, y, x+w, y, x+w, y+h, x, y+h}
end

function chanceList(...)
    return {
        chances = {},
        chance_defs = {...},
        next = function(self)
            -- Recreate list
            if #self.chances == 0 then
                for _, chance_def in ipairs(self.chance_defs) do
                    if type(chance_def) ~= 'table' then error('Wrong format. Tables must be passed in in the format: (value, number).') end
                    if type(chance_def[2]) ~= 'number' then error('Wrong format. Tables must be passed in in the format: (value, number).') end
                    -- Make sure that a false/nil value is converted to strings so that it doesn't break the table
                    if type(chance_def[2]) == 'boolean' then
                        if chance_def[2] == false then chance_def[2] = 'false' end
                    end
                    for i = 1, chance_def[2] do table.insert(self.chances, chance_def[1]) end
                end
            end
            -- Find value
            local value = table.remove(self.chances, love.math.random(1, #self.chances))
            if value == 'false' then value = nil end -- Convert false/nil value to boolean
            return value
        end,
    }
end

function iden(value)
    if type(value) == 'table' then return love.math.random(value[1], value[2])
    else return value end
end

function math.roundm(n, m)
    return math.floor(((n + m - 1)/m))*m
end

function math.sign(n)
    if n > 0 then return 1
    elseif n < 0 then return -1
    else return 0 end
end

uids = {}
function getUID(id)
    if id then
        if not uids[id] then
            uids[id] = true
            return id
        else return id end
    else
        local uid = Lume.uuid()
        while uids[uid] do uid = Lume.uuid() end
        uids[uid] = true
        return uid
    end
end

function isPointInsidePolygon(px, py, vertices)
    local nvert = #vertices/2
    local i, j = 1, 1
    local located_in_polygon = false

    for i = 1, nvert do
        if i == nvert then j = 1
        else j = i+1 end
        local vx1, vy1 = vertices[2*(i-1)+1], vertices[2*(i-1)+1+1]
        local vx2, vy2 = vertices[2*(j-1)+1], vertices[2*(j-1)+1+1]
        local below_low_y = vy1 > py
        local below_high_y = vy2 > py
        local within_ys_edges = below_low_y ~= below_high_y
        if within_ys_edges then
            local slope_of_line = (vx2-vx1)/(vy2-vy1)
            local point_on_line = (slope_of_line*(py-vy1))+vx1
            local is_left_to_line = px < point_on_line
            if is_left_to_line then
                located_in_polygon = not located_in_polygon
            end
        end
    end
    return located_in_polygon
end

function lineToGrid(element_index, grid_width, element_width, element_height, spacing_width, spacing_height)
    local x, y = element_index % grid_width, math.ceil(element_index/grid_width)
    if x == 0 then x = grid_width end
    return (x-1)*element_width + (x-1)*spacing_width, (y-1)*element_height + (y-1)*spacing_height
end
