local Area = Object:extend()

function Area:new(stage)
    self.stage = stage
    self.game_objects = {}
    self.world = Physics.newWorld({explicit_collision_events = true})
    setCollisionClasses(self)
end

function Area:update(dt)
    self.world:update(dt) 

    self.cells = {}
    self.cell_size = 96

    -- Set performance friendly data structures, general pre-update set
    for _, game_object in ipairs(self.game_objects) do
        -- Set game object spatial partition for fast retrieval
        local cell_x, cell_y = math.floor(game_object.x/self.cell_size), math.floor(game_object.y/self.cell_size)
        if not self.cells[cell_x] then self.cells[cell_x] = {} end
        if not self.cells[cell_x][cell_y] then self.cells[cell_x][cell_y] = {} end
        local c = self.cells[cell_x][cell_y]
        c[#c+1] = game_object
    end

    -- Update all objects
    for i = #self.game_objects, 1, -1 do
        local game_object = self.game_objects[i]
        game_object:update(dt)

        -- Remove dead objects
        if game_object.dead then
            game_object:destroy()
            table.remove(self.game_objects, i)
        end
    end
end

function Area:draw()
    for _, game_object in ipairs(self.game_objects) do game_object:draw() end
end

function Area:addGameObject(game_object_type, x, y, opts)
    local opts = opts or {}
    local game_object = _G[game_object_type](self, x or 0, y or 0, table.merge(opts, {class_name = game_object_type}))
    table.insert(self.game_objects, game_object)
    return game_object
end

function Area:getGameObject(id)
    for _, game_object in ipairs(self.game_objects) do
        if game_object.id == id then return game_object end
    end
end

function Area:destroy()
    for _, game_object in ipairs(self.game_objects) do game_object.dead = true end
    self:update(0)
    self.game_objects = {}
    self.world:destroy()
end

function Area:getAllGameObjectsThat(filter)
    local out = {}
    for _, game_object in pairs(self.game_objects) do
        if filter(game_object) then
            table.insert(out, game_object)
        end
    end
    return out
end

function Area:fastGameObjectQuery(x, y, radius)
    local out = {}
    local cell_x1, cell_y1 = math.floor((x - radius)/self.cell_size), math.floor((y - radius)/self.cell_size)
    local cell_x2, cell_y2 = math.floor((x + radius)/self.cell_size), math.floor((y + radius)/self.cell_size)
    for i = cell_x1, cell_x2 do
        for j = cell_y1, cell_y2 do
            local cell_x, cell_y = i, j
            if self.cells[cell_x] then
                local cell_objects = self.cells[cell_x][cell_y]
                if cell_objects then
                    for _, object in ipairs(cell_objects) do
                        local d = distance(x, y, object.x, object.y)
                        if d < radius then out[#out+1] = object end
                    end
                end
            end
        end
    end
    return out
end

function Area:getClosestGameObject(id, x, y, radius, object_types)
    local nearby_objects = self:fastGameObjectQuery(x, y, radius)
    nearby_objects = fn.select(nearby_objects, function(k, v) return fn.any(object_types, v.class_name) and v.id ~= id end)
    table.sort(nearby_objects, function(a, b) return distance(x, y, a.x, a.y) < distance(x, y, b.x, b.y) end)
    return nearby_objects[1]
end

function Area:queryCircleArea(x, y, radius, object_types)
    if not object_types then object_types = {} end
    return self:getAllGameObjectsThat(function(e)
        if fn.include(object_types, e.class_name) and 
           Math.polygon.getCircleIntersection(x, y, radius, rectangleToVertexList(e.x - e.w/2, e.y - e.h/2, e.w, e.h)) then
            return true
        end
    end)
end

function Area:queryRectangleArea(x, y, w, h, object_types)
    if not object_types then object_types = {} end
    return self:getAllGameObjectsThat(function(e)
        if fn.include(object_types, e.class_name) and 
           Math.polygon.isPolygonInside(rectangleToVertexList(x, y, w, h), rectangleToVertexList(e.x - e.w/2, e.y - e.h/2, e.w, e.h)) then
            return true
        end
    end)
end

function Area:queryPolygonArea(vertices, object_types)
    if not object_types then object_types = {} end
    return self:getAllGameObjectsThat(function(e)
        if fn.include(object_types, e.class_name) and 
           Math.polygon.isPolygonInside(vertices, e.vertices or rectangleToVertexList(e.x - e.w/2, e.y - e.h/2, e.w, e.h)) then
            return true
        end
    end)
end

function Area:queryLine(x1, y1, x2, y2, object_types)
    if not object_types then object_types = {} end
    return self:getAllGameObjectsThat(function(e)
        if fn.include(object_types, e.class_name) and 
           Math.polygon.getSegmentIntersection(x1, y1, x2, y2, rectangleToVertexList(e.x - e.w/2, e.y - e.h/2, e.w, e.h)) then
            return true
        end
    end)
end


return Area
